import os
from array import array
from pathlib import Path
from types import TracebackType
from typing import Dict, Union, Optional, Type, List

_Path = Union[str, bytes, os.PathLike[str]]
_Sequence = Union[str, bytes, bytearray, memoryview, array]

__version__: str
__author__: str
__build__: Dict[str, object]


class Hit:
    def __init__(self, identity: float, query_name: str, query_fraction: float, reference_name: str, reference_fraction: float) -> None: ...
    def __repr__(self) -> str: ...
    @property
    def identity(self) -> float: ...
    @property
    def query_name(self) -> str: ...
    @property
    def query_fraction(self) -> float: ...
    @property
    def reference_name(self) -> str: ...
    @property
    def reference_fraction(self) -> float: ...

class Sketch:
    @property
    def name(self) -> str: ...
    @property
    def amino_acid(self) -> bool: ...

class Database:
    @classmethod
    def load(cls, path: _Path) -> Database: ...
    @classmethod
    def open(cls, path: _Path) -> Database: ...

    def __init__(
        self,
        path: Union[str, bytes, os.PathLike[str], None] = None,
        *,
        compression: int = ...,
        marker_compression: int = ...,
        k: int = ...,
    ) -> None: ...

    def __enter__(self) -> Database: ...
    def __exit__(
        self, exc_type: Optional[Type[BaseException]], exc: Optional[BaseException], traceback: Optional[TracebackType]
    ) -> Optional[bool]: ...

    @property
    def path(self) -> Optional[Path]: ...

    def sketch(self, name: str, *contigs: _Sequence) -> None: ...
    def query(self, name: str, *contigs: _Sequence) -> List[Hit]: ...
        
    def save(self, path: _Path) -> None: ...
    def flush(self) -> None: ...


from . import _skani
from ._skani import Sketch, Database, Hit

__version__ = _skani.__version__
__author__ = _skani.__author__
__doc__ = _skani.__doc__
__build__ = _skani.__build__
__all__ = [
    "Sketch",
    "Database",
    "Hit",
    "SKANI_VERSION",
]

# Expose the version of embedded skani
SKANI_VERSION = _skani.__build__["dependencies"]["skani"]  # type: ignore

# Small addition to the docstring: show a link redirecting to the
# online version of the documentation, but this can only work when
# Python is running with docstrings enabled
if __doc__ is not None:
    __doc__ += """\nSee Also:
    An online rendered version of the documentation for this version
    of the library on
    `Read The Docs <https://{}.readthedocs.io/en/v{}/>`_.

    """.format(
        __name__,
        __version__
    )




from . import (
    test_ani,
    test_database
)

def load_tests(loader, suite, pattern):
    suite.addTests(loader.loadTestsFromModule(test_ani))
    suite.addTests(loader.loadTestsFromModule(test_database))
    return suite



import os
import pathlib
import tempfile
import unittest

import pyskani


class TestDatabase(unittest.TestCase):
    
    def test_memory(self):
        database = pyskani.Database()
        database.sketch("test genome", b"ATGC"*100)
        self.assertIs(database.path, None)

    def test_folder(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            database = pyskani.Database(tmpdir)
            database.sketch("test", b"ATGC"*100)
            self.assertTrue(os.path.exists(os.path.join(tmpdir, "test.sketch")))
            self.assertFalse(os.path.exists(os.path.join(tmpdir, "markers.bin")))
            database.flush()
            self.assertTrue(os.path.exists(os.path.join(tmpdir, "test.sketch")))
            self.assertTrue(os.path.exists(os.path.join(tmpdir, "markers.bin")))
            self.assertEqual(database.path, pathlib.Path(tmpdir))



import gzip
import os
import tempfile
import unittest

import pyskani
from . import fasta

# TODO: Update to use `pkg-resources`
DATA_FOLDER = os.path.realpath(os.path.join(__file__, os.pardir))


@unittest.skipUnless(os.path.exists(DATA_FOLDER), "missing data folder")
class TestAniEC590(unittest.TestCase):

    @staticmethod
    def load(filename):
        with gzip.open(os.path.join(DATA_FOLDER, f"{filename}.gz"), mode="rt") as f:
            return next(fasta.parse(f))

    @classmethod
    def setUpClass(cls):
        cls.ref = cls.load("e.coli-EC590.fasta")
        cls.db = pyskani.Database()
        cls.db.sketch("EC590", cls.ref.seq.encode("ascii"))
        cls.query = cls.load("e.coli-K12.fasta")

    def test_basic(self):
        hits = self.db.query("K12", self.query.seq.encode("ascii"))
        self.assertEqual(len(hits), 1)
        self.assertAlmostEqual(hits[0].reference_fraction, 0.9246, places=4)
        self.assertAlmostEqual(hits[0].query_fraction, 0.9189, places=4)
        self.assertAlmostEqual(hits[0].identity, 0.9939, places=4)

    def test_no_learned_ani(self):
        hits = self.db.query("K12", self.query.seq.encode("ascii"), learned_ani=False)
        self.assertEqual(len(hits), 1)
        self.assertAlmostEqual(hits[0].reference_fraction, 0.9246, places=4)
        self.assertAlmostEqual(hits[0].query_fraction, 0.9189, places=4)
        self.assertAlmostEqual(hits[0].identity, 0.9946, places=4)

    def test_learned_ani(self):
        hits = self.db.query("K12", self.query.seq.encode("ascii"), learned_ani=True)
        self.assertEqual(len(hits), 1)
        self.assertAlmostEqual(hits[0].reference_fraction, 0.9246, places=4)
        self.assertAlmostEqual(hits[0].query_fraction, 0.9189, places=4)
        self.assertAlmostEqual(hits[0].identity, 0.9939, places=4)

    def test_robust(self):
        hits = self.db.query("K12", self.query.seq.encode("ascii"), robust=True)
        self.assertEqual(len(hits), 1)
        self.assertAlmostEqual(hits[0].reference_fraction, 0.9246, places=4)
        self.assertAlmostEqual(hits[0].query_fraction, 0.9189, places=4)
        self.assertAlmostEqual(hits[0].identity, 0.9978, places=4)

    def test_median(self):
        hits = self.db.query("K12", self.query.seq.encode("ascii"), median=True)
        self.assertEqual(len(hits), 1)
        self.assertAlmostEqual(hits[0].reference_fraction, 0.9246, places=4)
        self.assertAlmostEqual(hits[0].query_fraction, 0.9189, places=4)
        self.assertAlmostEqual(hits[0].identity, 0.9995, places=4)


import collections


class Record(collections.namedtuple("Record", ["id", "seq", "description"])):
    pass


def parse(path):

    if isinstance(path, str):
        file = open(path)
    else:
        file = path

    with file:

        id_ = None
        seq = []
        for line in file:
            l = line.strip()
            if line.startswith(">"):
                if id_ is not None:
                    yield Record(id_, "".join(seq), desc)
                id_ = line[1:].split()[0].strip()
                desc = " ".join(line[1:].split(maxsplit=1))
                seq = []
            elif l:
                seq.append(l)
        if id_ is not None:
            yield Record(id_, "".join(seq), desc)
        elif seq:
            raise ValueError("not in FASTA format")



#!/usr/bin/env python3

import configparser
import os
import shutil
import subprocess
import sys
import urllib.request
from distutils.errors import DistutilsPlatformError
from distutils.log import INFO

import setuptools
import setuptools_rust as rust
from setuptools.command.sdist import sdist as _sdist
from setuptools_rust.build import build_rust as _build_rust

try:
    from setuptools_rust.rustc_info import get_rust_version
except ImportError:
    from setuptools_rust.utils import get_rust_version


class vendor(setuptools.Command):

    description = "vendor Rust dependencies into a local folder"
    user_options = [
        ("vendor-dir=", "d", "the path where to vendor the Rust crates")
    ]

    def initialize_options(self):
        self.vendor_dir = None

    def finalize_options(self):
        if self.vendor_dir is None:
            self.vendor_dir = "crates"

    def run(self):
        # make sure rust is available
        _build_cmd = self.get_finalized_command("build_rust")
        rustc = get_rust_version()
        if rustc is None:
            _build_cmd.setup_temp_rustc_unix(toolchain="stable", profile="minimal")
        # vendor crates
        proc = subprocess.run(["cargo", "vendor", self.vendor_dir])
        proc.check_returncode()
        # write the cargo config file
        self.mkpath(".cargo")
        with open(os.path.join(".cargo", "config.toml"), "w") as f:
            f.write(
                """
                [source.crates-io]
                replace-with = "vendored-sources"

                [source.vendored-sources]
                directory = "{}"
                """.format(self.vendor_dir)
            )


class sdist(_sdist):

    def run(self):
        # build `pyproject.toml` from `setup.cfg`
        c = configparser.ConfigParser()
        c.add_section("build-system")
        c.set("build-system", "requires", str(self.distribution.setup_requires))
        c.set("build-system", 'build-backend', '"setuptools.build_meta"')
        with open("pyproject.toml", "w") as pyproject:
            c.write(pyproject)

        # run the rest of the packaging
        _sdist.run(self)


class build_rust(_build_rust):

    def run(self):
        rustc = get_rust_version()
        if rustc is not None:
            nightly = rustc is not None and "nightly" in rustc.prerelease
        else:
            self.setup_temp_rustc_unix(toolchain="stable", profile="minimal")
            nightly = False

        if self.inplace:
            self.extensions[0].strip = rust.Strip.No
        if nightly:
            self.extensions[0].features = (*self.extensions[0].features, "nightly")

        _build_rust.run(self)


    def setup_temp_rustc_unix(self, toolchain, profile):
        rustup_sh = os.path.join(self.build_temp, "rustup.sh")
        os.environ["CARGO_HOME"] = os.path.join(self.build_temp, "cargo")
        os.environ["RUSTUP_HOME"] = os.path.join(self.build_temp, "rustup")

        self.mkpath(os.environ["CARGO_HOME"])
        self.mkpath(os.environ["RUSTUP_HOME"])

        self.announce("downloading rustup.sh install script", level=INFO)
        with urllib.request.urlopen("https://sh.rustup.rs") as res:
            with open(rustup_sh, "wb") as dst:
                shutil.copyfileobj(res, dst)

        self.announce("installing Rust compiler to {}".format(self.build_temp), level=INFO)
        proc = subprocess.run([
            "sh",
            rustup_sh,
            "-y",
            "--default-toolchain",
            toolchain,
            "--profile",
            profile,
            "--no-modify-path"
        ])
        proc.check_returncode()

        self.announce("updating $PATH variable to use local Rust compiler", level=INFO)
        os.environ["PATH"] = ":".join([
            os.path.abspath(os.path.join(os.environ["CARGO_HOME"], "bin")),
            os.environ["PATH"]
        ])


    def get_dylib_ext_path(self, ext, module_name):
        ext_path = _build_rust.get_dylib_ext_path(self, ext, module_name)
        if self.inplace:
            base = os.path.basename(ext_path)
            folder = os.path.dirname(os.path.realpath(__file__))
            prefix = os.path.sep.join(ext.name.split(".")[:-1])
            ext_path = os.path.join(folder, prefix, base)
        return ext_path


# HACK: Use the `configparser` from Python to read the `Cargo.toml`
#       manifest file (this... works) so that the package version
#       can be extracted from there and synchronized everywhere.
parser = configparser.ConfigParser()
parser.read(os.path.join(os.path.dirname(__file__), "Cargo.toml"))
version = parser.get("package", "version").strip('"')

setuptools.setup(
    version=version,
    setup_requires=["setuptools", "setuptools_rust"],
    cmdclass=dict(sdist=sdist, build_rust=build_rust, vendor=vendor),
    rust_extensions=[rust.RustExtension(
        "pyskani._skani",
        path="Cargo.toml",
        binding=rust.Binding.PyO3,
        strip=rust.Strip.Debug,
        features=["extension-module"],
    )],
)
